-- Required disclosure, as per ZeniMax Media Inc.
--    Further details at: https://account.elderscrollsonline.com/add-on-terms

-- This Add-on is not created by, affiliated with or sponsored by ZeniMax Media
-- Inc. or its affiliates. The Elder Scrolls® and related logos are registered
-- trademarks or trademarks of ZeniMax Media Inc. in the United States and/or
-- other countries. All rights reserved.



-- GrayMarket is a simple guild market pricing add-on which displays item sales
--    stats, where available, on the hovered item tooltip with a breakdown
--    by each guild the player is a member of

-- Version: 1.0.4
-- Created: June 1, 2015
-- Updated: June 22, 2015

-- Copyright (c) 2015, Mike Gauthier (@gray.malken)
-- All rights reserved.
-- encl: 'license' text file

-- namespace for the market
local GrayMarket = GrayMarket or {}

-- constants, defaults, and static keys
local LAST_FETCH = "lastFetch"
local GUILD_LIST = "list"
local GUILD_SALE_LIST = "sales"
local VENDOR_SALE = "vendor"
local MEMBER_SALE = "member"
local GUILD_MEMBERS = "guildMembers"
local TRANSACTION_TIME = "tTime"
local PRICE_LIST = "priceList"
local ITEM_QUANTITY = "qty"
local ITEM_PRICE = "price"
local TOTAL_QUANTITY = "tQty"
local TOTAL_PRICE = "tPrice"
local AVERAGE_PRICE = "avg"
local TOTAL_GUILDS = "totalGuilds"
local CALL_DELAY = 5000
local CONTINUOUS_UPDATE_INTERVAL = 900000 -- 300000 --
local DAY_S = 86400
local DEFAULT_DATA_WINDOW = 15
local DEFAULT_MAX_WINDOW_S = DEFAULT_DATA_WINDOW * DAY_S
local SHOW_ALL_GUILDS = "settingShowAllGuilds"
local RUN_AUTOMATICALLY = "settingRunAutomatically"
local SALE_WINDOW = "settingDataWindow"
local FRESH_DATA = {
  [GUILD_LIST] = {},
  [PRICE_LIST] = {},
  [LAST_FETCH] = 0,
  [SALE_WINDOW] = DEFAULT_MAX_WINDOW_S,
  [SHOW_ALL_GUILDS] = true,
  [RUN_AUTOMATICALLY] = false
}

-- default icons, custom fonts and layout template for on-mouseover
local ICON_VENDOR = "|t24:24:esoui/art/guild/guild_tradingHouseAccess.dds|t"
local ICON_MEMBER = "|t20:20:esoui/art/guild/guild_heraldryaccess.dds|t"
local ICON_GOLD = "|t14:14:EsoUI/Art/currency/currency_gold.dds|t"

local FONT_ITEM = "$(BOLD_FONT)|14|soft-shadow-thin"
local TEXT_DEFAULT_DATA = "|u1:1:text:(%s SOLD):|u%s " .. ICON_GOLD .. " AVG"

local HEADING_DEFAULT = "|u1:1:text:(# SOLD):|u " .. ICON_GOLD .. " AVG"
local HEADING_VENDOR = "Vender " .. ICON_VENDOR
local HEADING_MEMBER = "Member " .. ICON_MEMBER
local HEADING_FULL = HEADING_VENDOR .. " or  " .. HEADING_MEMBER .. HEADING_DEFAULT

-- formatting constaint on-mouseover quantities
local VALUE_CAP = 100000

-- settings panel and options via LibAddonMenu-2.0
local PANEL_DATA = {
  type = "panel",
  name = "GrayMarket",
  author = "@gray.malken",
  version = "1.0.4",
  registerForRefresh = true,
  registerForDefaults = true
}

local OPTIONS_TABLE = {
  [1] = {
    type = "description",
		text = "Running GrayMarket automatically is best if this is the only market add-on installed. By default, GrayMarket does not update sales data unless you trigger it via these settings or your own keybind. You may have ESO API message limit errors if other market add-ons are installed and GrayMarket is running in automatic mode.",
		width = "full"
  },
  [2] = {
    type = "checkbox",
  	name = "Run automatically?",
  	tooltip = "Check to run automatically on startup, and in the background.",
  	getFunc = function() return GrayMarket.data[RUN_AUTOMATICALLY] end,
  	setFunc = function(value) GrayMarket.data[RUN_AUTOMATICALLY] = value end,
  	width = "full",
  	warning = "UI reload required.",
  	default = false
  },
  [3] = {
    type = "checkbox",
    name = "Show data if guild store is still locked?",
    tooltip = "Check to show data on item tooltip for guilds without guild store enabled.",
    getFunc = function() return GrayMarket.data[SHOW_ALL_GUILDS] end,
    setFunc = function(value) GrayMarket.data[SHOW_ALL_GUILDS] = value end,
    width = "full",
    default = true
  },
  [4] = {
    type = "slider",
		name = "Size of data history in days?",
		tooltip = "The number of days to search for and keep sales data on, up to today.",
		min = 5,
		max = 60,
		step = 5,
		getFunc = function() return GrayMarket:GetDataWindow() end,
		setFunc = function(value) GrayMarket:SetDataWindow(value) end,
		width = "full",
		default = DEFAULT_DATA_WINDOW
  },
  [5] = {
    type = "button",
    name = "Fetch and Purge",
    tooltip = "Perform a full fetch and purge of data, typical of a full GrayMarket startup.",
    func = function() GrayMarket:PerformFullPurgeAndFetch() end,
    width = "half"
  }
}

-- libs
local AOM = LibStub:GetLibrary("LibAddonMenu-2.0")

-- ad-hoc strings (IF ever localizing, move appropriately)
ZO_CreateStringId("SI_BINDING_NAME_GRAYMARKET_FETCH", "Manually Fetch Sales")

-- variables
GrayMarket.name = "GrayMarket"
GrayMarket.dataVersion = 1
GrayMarket.numGuilds = 0
GrayMarket.guildPasses = {}
GrayMarket.fetchInProgress = false
GrayMarket.purging = false
GrayMarket.newSales = 0
GrayMarket.data = FRESH_DATA
GrayMarket.tooltipVisible = false
GrayMarket.itemLink = nil
GrayMarket.summary = {
  [TOTAL_GUILDS] = 0,
  [TOTAL_QUANTITY] = 0,
  [TOTAL_PRICE] = 0
}

--[[
  Initialize():

    Perform add-on and data initialization operations.
--]]
function GrayMarket:Initialize()
  -- setup options/settings
  AOM:RegisterAddonPanel(GrayMarket.name, PANEL_DATA)
  AOM:RegisterOptionControls(GrayMarket.name, OPTIONS_TABLE)

  -- how many guilds you in?
  self.numGuilds = GetNumGuilds()

  -- load saved data
  self.data = ZO_SavedVars:NewAccountWide("GrayMarketData", self.dataVersion, nil, FRESH_DATA, "Default")

  -- item hover
  ZO_PreHookHandler(ItemTooltip, "OnUpdate", function() self:OnTooltipUpdate() end)
  ZO_PreHookHandler(ItemTooltip, "OnCleared", function() self:OnTooltipCleared() end)

  -- get new sales data
  if self.data[RUN_AUTOMATICALLY] then
    zo_callLater(function() self:PerformFullPurgeAndFetch() end, CALL_DELAY)
  end
end

--[[
  SetDataWindow(windowInDays):

    Take supplied days and convert to seconds for the saved setting sale/data
    window.

    @param windowInDays   The number of days worth of data to keep
]]--
function GrayMarket:SetDataWindow(windowInDays)
  self.data[SALE_WINDOW] = windowInDays * DAY_S
end

--[[
  GetDataWindow():

    Return the sale window in days.  Used internally as seconds.
]]--
function GrayMarket:GetDataWindow()
  return self.data[SALE_WINDOW] / DAY_S
end

--[[
  LinkToChat():

    Insert a sales summary to chat based on current item
]]--
function GrayMarket:LinkToChat()
  if self.itemLink then
    local textEntry = self.name .. ": Last " .. self:GetDataWindow() .. " days " .. string.gsub(self.itemLink, '|H0', '|H1') .. " %d sold, %d guilds: %dg AVG"
    local average = 0

    if self.summary[TOTAL_QUANTITY] > 0 then
      average = math.ceil(self.summary[TOTAL_PRICE] / self.summary[TOTAL_QUANTITY])
    end

    StartChatInput()
    ZO_ChatWindowTextEntryEditBox:InsertText(string.format(textEntry, self.summary[TOTAL_QUANTITY], self.summary[TOTAL_GUILDS], average))
  end
end

--[[
  PurgeOldGuilds():

    Remove sales data from guilds not currently a member of
]]--
function GrayMarket:PurgeOldGuilds()
  -- ensure guild data matches guilds the player belongs to
  for key, value in pairs(self.data[GUILD_LIST]) do
    local guildName = nil

    for guildId = 1, self.numGuilds do
      guildName = GetGuildName(guildId)

      if key == guildName then
        break
      else
        guildName = nil
      end
    end

    -- drop from GUILD_LIST
    if guildName == nil then
      self.data[GUILD_LIST][key] = nil
    end
  end
end

--[[
  PurgeOldSales():

    Remove sales data from storage that falls outside of the defined sale
    window in add-on settings.
]]--
function GrayMarket:PurgeOldSales()
  -- for reporting to player
  local numberPurged = 0

  -- purge data outside the configured sale window for each guild
  for guildId = 1, self.numGuilds do
    local guildName = GetGuildName(guildId)

    if not self.data[GUILD_LIST][guildName] then
      break
    end

    for gIndex, gElements in pairs(self.data[GUILD_LIST][guildName]) do
      -- checks guild level element for the sales list
      if gIndex == GUILD_SALE_LIST then
        -- moves through the type of sale, vendor/member
        for saleType, items in pairs(gElements) do
          -- each salesheet contains sales by item ID
          for itemIndex, saleSheet in pairs(items) do
            -- the receipt contains the buyer/seller information
            for i, receipt in pairs(saleSheet) do
              if receipt[TRANSACTION_TIME] < (GetTimeStamp() - self.data[SALE_WINDOW]) then
                -- drop from the sales list
                table.remove(saleSheet, i)
              end
            end

            if table.getn(items[itemIndex]) == 0 then
              items[itemIndex] = nil
            end
          end
        end
      end
    end
  end

  if numberPurged > 0 then
    d(self.name .. ": Purged " .. numberPurged .. " sale(s) older than " .. (self.data[SALE_WINDOW] / DAY_S) .. " days.")
  end
end

--[[
  RebuildGuildRoster():

    Update the list of guild members in storage, per guild.
--]]
function GrayMarket:RebuildGuildRoster()
  -- iterate player guilds by index, but store by name (indexes can change)
  for guildId = 1, self.numGuilds do
    local guildName = GetGuildName(guildId)

    -- test for 'fresh' build
    if not self.data[GUILD_LIST][guildName] then
      self.data[GUILD_LIST][guildName] = {}
    end

    -- reset roster for each guild in storage
    local guildData = self.data[GUILD_LIST][guildName]
    guildData[GUILD_MEMBERS] = {}

    for i = 1, GetNumGuildMembers(guildId) do
      local name = GetGuildMemberInfo(guildId, i)
      guildData[GUILD_MEMBERS][name] = {}
    end
  end
end

--[[
  Purge():

    Purge old guild data (not a member anymore, perhaps?) and old sales data,
    like sales outside of application data window
]]--
function GrayMarket:Purge()
  self:PurgeOldGuilds()
  self:PurgeOldSales()

  -- rebuild roster for each guild for if members leave, etc
  self:RebuildGuildRoster()

  self.purging = false
end

--[[
  PerformFullPurgeAndFetch():

    Entry for a full forced fetch which includes old data purge.  typical
    of add-on startup in auto mode, or manual triggers.
]]--
function GrayMarket:PerformFullPurgeAndFetch()
  if not self.purging and not self.fetchInProgress then
    self.purging = true

    -- clean up old or out-dated data
    zo_callLater(function() self:Purge() end, CALL_DELAY)

    -- stagger kick off of sales fetch
    zo_callLater(function() self:GetAllSales() end, (CALL_DELAY * 2))

    -- warn we are in progress
    d(self.name .. ": Fetching sales, purging old data...")
  else
    -- bail if already working on anything
    d(self.name .. ": Sales fetch is already in progress...")
  end
end

--[[
  GetAllSales():

    Fetch sales history, typically on startup. Uses
    EVENT_GUILD_HISTORY_RESPONSE_RECEIVED event registration. For each guild,
    results on callback shall be handled via
    'GrayMarket.OnResponseReceived(event, guildId, category)'.

    While a request to 'GetAllSales()' is running, no concurrent request to
    fetch guild sales should be allowed. ie: when GrayMarket.fetchInProgress == true
--]]
function GrayMarket:GetAllSales()
  -- do not attempt until any running purge is complete
  if self.purging then
    zo_callLater(function() self:GetAllSales() end, CALL_DELAY)
  end

  -- bail if already working on sales data
  if self.fetchInProgress then
    d(self.name .. ": Sales fetch is already in progress...")
    return
  end

  self.fetchInProgress = true

  -- price monitoring callback
  EVENT_MANAGER:RegisterForEvent(self.name, EVENT_GUILD_HISTORY_RESPONSE_RECEIVED, self.OnResponseReceived)

  -- get new sales for each guild
  self:GetNewSalesHistory()
end

--[[
  GetNewSalesHistory():

    Used to continuously get new sales history, if available, at an interval
    of CONTINUOUS_UPDATE_INTERVAL, or 15 minutes by default should
    GrayMarket.data[RUN_AUTOMATICALLY] be set true

]]--
function GrayMarket:GetNewSalesHistory()
  if self.numGuilds > 0 then
    -- reset progress and related
    self.fetchInProgress = true
    self.guildPasses = {}

    for i = 1, self.numGuilds do
      zo_callLater(function() self:GetSalesHistory(GetGuildId(i), true) end, CALL_DELAY)
    end

    -- poll for when done
    zo_callLater(function() self:CheckDone() end, CALL_DELAY * self.numGuilds)
  else
    d(self.name .. ": Please join some guilds first!  GrayMarket can't work without them, AND, they're kinda fun!")
    self.fetchInProgress = false
  end

  -- setup next full update
  if self.data[RUN_AUTOMATICALLY] then
    zo_callLater(function() self:GetNewSalesHistory() end, CONTINUOUS_UPDATE_INTERVAL)
  end
end

--[[
  CheckDone():

    Test, typically via polling mechanism in this system, if the guild history
    pertaining to guild sales has been retrieved, old data purged, and new
    formatted and saved.

    This test checks if a pass to retrieve newer sales data and older data
    has been completed via the GetSalesHistory() method and the API methods
    'RequestGuildHistoryCategoryNewest' and 'RequestGuildHistoryCategoryOlder'
    specifically.
--]]
function GrayMarket:CheckDone()
  -- test if all sales have been Updated
  for key, val in pairs(self.guildPasses) do
    if val ~= 2 then
      -- poll for when done until we are
      zo_callLater(function() self:CheckDone() end, CALL_DELAY)
      return
    end
  end

  -- process price list
  self:UpdatePriceList()
  self.fetchInProgress = false

  d(self.name .. ": Updated with " .. self.newSales .. " new sales.")

  -- clean up
  self.newSales = 0
  self.data[LAST_FETCH] = GetTimeStamp()
end

--[[
  OnTooltipCleared():

    Event handler for when an ItemTooltip is cleared away after being shown
]]--
function GrayMarket:OnTooltipCleared()
  GrayMarket.tooltipVisible = false

  self.itemLink = nil
  self.summary = {
    [TOTAL_GUILDS] = 0,
    [TOTAL_QUANTITY] = 0,
    [TOTAL_PRICE] = 0
  }
end

--[[
  OnTooltipUpdate():

    Event handler for when an ItemTooltip is shown to the player, typically
    on mouse hover for an item in inventory, bank, merchant, etc
]]--
function GrayMarket:OnTooltipUpdate()
  if self.tooltipVisible then
    return
  end

  -- data concerning what the player is mousing over
  local moc = moc()
  local itemLink = nil
  self.tooltipVisible = true

  if not moc.slotType and moc.dataEntry then
    local data = moc.dataEntry.data

    if data.sellerName then
      local myName = GetDisplayName()

      -- some add-ons modify sellername and attach extra information ..
      -- look for whether it's there, not a perfect match
      if data.sellerName == "" or string.find(data.sellerName, myName) then
        -- has been seen as empty on occasion .. something to try and keep track of?
        itemLink = GetTradingHouseListingItemLink(data.slotIndex)
      else
        if data.name ~= "Guild Tabard" then
          -- search results from guild store
          itemLink = GetTradingHouseSearchResultItemLink(data.slotIndex)
        end
      end
    elseif data.bagId and data.slotIndex then
      -- covers almost all listed items
      itemLink = GetItemLink(data.bagId, data.slotIndex)
    end
  else
    if moc.bagId and moc.slotIndex then
      -- anything in a bag or bank slot
      itemLink = GetItemLink(moc.bagId, moc.slotIndex)
    elseif moc.slotType == SLOT_TYPE_MAIL_ATTACHMENT then
      -- for player attachments
      itemLink = GetAttachedItemLink(MAIL_INBOX:GetOpenMailId(), moc.id, LINK_STYLE_DEFAULT)
    elseif moc.slotType == SLOT_TYPE_MAIL_QUEUED_ATTACHMENT then
      -- for attaching things to other players
      itemLink = GetMailQueuedAttachmentLink(moc.id, LINK_STYLE_DEFAULT)
    elseif moc.slotType == SLOT_TYPE_LOOT then
      -- loot window
      itemLink = GetLootItemLink(moc.dataEntry.data.lootId)
    elseif moc.slotType == SLOT_TYPE_STORE_BUY then
      -- store listing
      itemLink = GetStoreItemLink(moc.index)
    elseif moc.slotType == SLOT_TYPE_STORE_BUYBACK then
      -- buyback listing
      itemLink = GetBuybackItemLink(moc.index)
    end
  end

  if not itemLink then
    self.tooltipVisible = false
    return
  end

  local itemId = GrayMarket:GetItemId(itemLink)

  if not itemId then
    self.tooltipVisible = false
    return
  end

  local itemPrices = self:LookupPrices(itemId)

  if itemPrices == {} then
    self.tooltipVisible = false
    return
  end

  self.itemLink = itemLink
  self:UpdateTooltip(itemId, itemPrices)
end

--[[
  AddEmptyGuildData(guildId):

    Add a line to ItemTooltip and colour code as "no data" or "guild store not enabled"

    @param guildId          The guild index without data
]]--
function GrayMarket:AddEmptyGuildData(guildId)
  if DoesGuildHavePrivilege(guildId, GUILD_PRIVILEGE_TRADING_HOUSE) then
    self:AddLineSummary(0, 0, 0, 0, 1, 0, 0)
  else
    -- takes note of user setting
    if GrayMarket.data[SHOW_ALL_GUILDS] then
      self:AddLineSummary(0, 0, 0, 0, 0.4, 0.4, 0.4)
    end
  end
end

function GrayMarket:AddGuildName(guildId, guildName)
  if DoesGuildHavePrivilege(guildId, GUILD_PRIVILEGE_TRADING_HOUSE) then
    ItemTooltip:AddLine(guildName, "ZoFontGameOutline")
    self.summary[TOTAL_GUILDS] = self.summary[TOTAL_GUILDS] + 1
  else
    -- takes note of user setting
    if GrayMarket.data[SHOW_ALL_GUILDS] then
      ItemTooltip:AddLine(guildName, "ZoFontGameOutline")
    end
  end
end

--[[
  AddLineSummary(numItemsVendor, averageValueVendor, numItemsMember, averageValueMember, red, green, blue):

    Base method for adding lines to ItemTooltip in GrayMarket format

    @param numItemsVendor           Number of sales from vendor sales type
    @param averageValueVendor       Average sale amount from vendor sales type
    @param numItemsMember           Number of sales from member sales type
    @param averageValueMember       Average sale amount from member sales type
    @param red                      Colour value as (255 / int colour), ie: 0.4
    @param green                    Colour value as (255 / int colour), ie: 0.4
    @param blue                     Colour value as (255 / int colour), ie: 0.4
]]--
function GrayMarket:AddLineSummary(numItemsVendor, averageValueVendor, numItemsMember, averageValueMember, red, green, blue)
  ItemTooltip:AddLine(
        ICON_VENDOR ..
        string.format(TEXT_DEFAULT_DATA,
            numItemsVendor,
            averageValueVendor
        ) ..
        "   " ..
        ICON_MEMBER ..
        string.format(TEXT_DEFAULT_DATA,
            numItemsMember,
            averageValueMember
        ),
        FONT_ITEM, red, green, blue)
end

--[[
  UpdateTooltip(itemId, itemPrices):

    Update the ItemTooltip with any available sales data for the supplied item

    @param itemId       The item to lookup pricing information for, and display
    @param itemPrices   The item pricing table with any available data for
                        supplied itemId
]]--
function GrayMarket:UpdateTooltip(itemId, itemPrices)
  -- add GrayMarket badge to tooltip
    ItemTooltip:AddVerticalPadding(20)
    ZO_Tooltip_AddDivider(ItemTooltip)
    ItemTooltip:AddLine(self.name, "ZoFontWinH2", 0.7647, 0.7647, 0.6078)
    ItemTooltip:AddLine(HEADING_FULL, FONT_ITEM, 0.7647, 0.7647, 0.6078)
    ZO_Tooltip_AddDivider(ItemTooltip)
    ItemTooltip:AddVerticalPadding(5)

    -- no data for this item in any guild
    if itemPrices[itemId] == nil then
      for guildId = 1, self.numGuilds do
        local guildName = GetGuildName(guildId)

        self:AddGuildName(guildId, guildName)
        self:AddEmptyGuildData(guildId)
      end

      return
    end

    for guildId = 1, self.numGuilds do
      local guildName = GetGuildName(guildId)
      self:AddGuildName(guildId, guildName)

      if itemPrices[itemId][guildName] then
        self:AddSaleSummary(guildId, itemPrices[itemId][guildName])
      else
        -- no pricing data for this guild
        self:AddEmptyGuildData(guildId)
      end
    end
end

--[[
  AddSaleSummary(guildId, dataItem):

    Add a line of sale summary to ItemTooltip for vendor and member sales,
    as the data is available.  Using this method assumes one or both sales
    types have data associated with them

    @param guildId    The guild index with sales data
    @param dataItem   The vendor/member sales data table
]]--
function GrayMarket:AddSaleSummary(guildId, dataItem)
  if not dataItem then
    self:AddEmptyGuildData(guildId)
    return
  end

  local uncappedTotalSold = dataItem[VENDOR_SALE][TOTAL_QUANTITY] or 0 + dataItem[MEMBER_SALE][TOTAL_QUANTITY] or 0
  self.summary[TOTAL_QUANTITY] = self.summary[TOTAL_QUANTITY] + uncappedTotalSold

  local uncappedTotalPrice = dataItem[VENDOR_SALE][TOTAL_PRICE] or 0 + dataItem[MEMBER_SALE][TOTAL_PRICE] or 0
  self.summary[TOTAL_PRICE] = self.summary[TOTAL_PRICE] + uncappedTotalPrice

  local numItemsVendor = self:GetCappedValue(dataItem[VENDOR_SALE][TOTAL_QUANTITY] or 0)
  local averageValueVendor = dataItem[VENDOR_SALE][AVERAGE_PRICE] or 0
  local numItemsMember = self:GetCappedValue(dataItem[MEMBER_SALE][TOTAL_QUANTITY] or 0)
  local averageValueMember = dataItem[MEMBER_SALE][AVERAGE_PRICE] or 0

  self:AddLineSummary(numItemsVendor, averageValueVendor, numItemsMember, averageValueMember, 0.7647, 0.7647, 0.6078)
end

function GrayMarket:GetCappedValue(suppliedValue)
  if suppliedValue > VALUE_CAP then
    return ">" .. VALUE_CAP
  end

  return suppliedValue
end

--[[
  LookupPrices(itemId):

    For a given itemId, typically being mouse-hovered, lookup any available
    pricing information by guild and sales type (vendor/member)

    @param itemId   The item identifier pulled from an item link and used
                    to lookup sales data we have stored, where itemId is a key
]]--
function GrayMarket:LookupPrices(itemId)
  local itemPrices = {}

  if not itemId or self.data[PRICE_LIST] == itemPrices then
    return itemPrices
  end

  -- prep lookup and result data
  local priceList = self.data[PRICE_LIST]

  -- get prices for this item by guild name
  for guildName, sales in pairs(priceList) do
    -- each price is further broken into vendor or member sales types
    for saleType, saleItems in pairs(sales) do
      for itemIndex, itemValues in pairs(saleItems) do
        if itemIndex == itemId then
          if not itemPrices[itemId] then
            itemPrices[itemId] = {}
          end

          if not itemPrices[itemId][guildName] then
            itemPrices[itemId][guildName] = {
              [MEMBER_SALE] = {},
              [VENDOR_SALE] = {}
            }
          end

          if priceList[guildName][saleType] and priceList[guildName][saleType][itemId] then
            itemPrices[itemId][guildName][saleType] = priceList[guildName][saleType][itemId]
          end
        end
      end
    end
  end

  return itemPrices
end

--[[
  UpdatePriceList():

    Update pricing totals on a per item basis, where player can see avg
    sales price by guild, sales type, and the number sold
]]--
function GrayMarket:UpdatePriceList()
  -- start a new list
  local priceList = {}

  for i = 1, self.numGuilds do
    local guildName = GetGuildName(i)
    priceList[guildName] = {
      [MEMBER_SALE] = {},
      [VENDOR_SALE] = {}
    }

    for gIndex, gElements in pairs(self.data[GUILD_LIST][guildName]) do
      -- checks guild level element for the sales list
      if gIndex == GUILD_SALE_LIST then
        -- moves through the type of sale, vendor/member
        for saleType, items in pairs(gElements) do
          -- each salesheet contains sales by item ID
          for itemIndex, saleSheet in pairs(items) do
            -- create sales type entry as necessary (vendor/member)
            if not priceList[guildName][saleType][itemIndex] then
              priceList[guildName][saleType][itemIndex] = {
                [TOTAL_QUANTITY] = 0,
                [TOTAL_PRICE] = 0,
                [AVERAGE_PRICE] = 0
              }
            end

            local itemTotals = priceList[guildName][saleType][itemIndex]

            -- the receipt contains the buyer/seller information
            for i, receipt in pairs(saleSheet) do
              itemTotals[TOTAL_QUANTITY] = itemTotals[TOTAL_QUANTITY] + receipt[ITEM_QUANTITY]
              itemTotals[TOTAL_PRICE] = itemTotals[TOTAL_PRICE] + receipt[ITEM_PRICE]
            end

            -- new average per item for this sale type, in this guild
            itemTotals[AVERAGE_PRICE] = math.ceil(itemTotals[TOTAL_PRICE] / itemTotals[TOTAL_QUANTITY])
          end
        end
      end
    end
  end

  -- swap to new list
  self.data[PRICE_LIST] = priceList
end

--[[
  GetSalesHistory(guildId, firstPass):

    Test, typically via polling mechanism in this system, if the guild history
    pertaining to guild sales has been retrieved, old data purged, and new
    formatted and saved.

    This test checks if a pass to retrieve newer sales data and older data
    has been completed via the GetSalesHistory() method and the API methods
    'RequestGuildHistoryCategoryNewest' and 'RequestGuildHistoryCategoryOlder'
    specifically.

    @param guildId    The guild index to use for making the request
    @param firstPass  A boolean value indicating true when making 'newest'
                      history requests and false for 'older' history
--]]
function GrayMarket:GetSalesHistory(guildId, firstPass)
  -- get newest sales for supplied guildId
  if firstPass then
    if not RequestGuildHistoryCategoryNewest(guildId, GUILD_HISTORY_STORE) then
      -- d("Unable to find sales data for guild: " .. GetGuildName(guildId))
    end

    return
  end

  -- when firstPass is false, an older history should be checked
  self.guildPasses[GetGuildName(guildId)] = 2
  RequestGuildHistoryCategoryOlder(guildId, GUILD_HISTORY_STORE)
end

--[[
  InitItem(guildName, saleType, itemId):

    Initialize the full path for an item in the data store (convenience method)
]]--
function GrayMarket:InitItem(guildName, saleType, itemId)
  if not self.data[GUILD_LIST][guildName][GUILD_SALE_LIST][saleType][itemId] then
    GrayMarket.data[GUILD_LIST][guildName][GUILD_SALE_LIST][saleType][itemId] = {}
  end
end

--[[
  GetItemId(itemLink):

    Get the item identifier from a supplied game item link

    @param itemLink   The item link string generated from various systems in-game
]]--
function GrayMarket:GetItemId(itemLink)
  local elements = {}

  -- for splitting out the item ID element, position 3
  for element in string.gmatch(itemLink or "", "(%d-):") do
    table.insert(elements, element)
  end

  if table.getn(elements) >= 3 then
    return elements[3]
  end

  return nil
end

--[[
  OnAddOnLoaded(event, addonName):

    Callback method registered for EVENT_ADD_ON_LOADED with the system
    EVENT_MANAGER.

    @param event      An event code attached to the notification
    @param addonName  The name of the add-on which fired this even.  When
                      this add-on, perform initialization
--]]
function GrayMarket.OnAddOnLoaded(event, addonName)
  if addonName == GrayMarket.name then
    GrayMarket:Initialize()
  end
end

--[[
  OnResponseReceived(event, guildId, category):

    Callback method registered for EVENT_GUILD_HISTORY_RESPONSE_RECEIVED with
    the system EVENT_MANAGER.  Used to trigger a parse of guild history looking
    specifically for GUILD_HISTORY_STORE events. Only sales since the last
    fetch will be factored into the system.

    @param event      An event code attached to the notification
    @param guildId    The guild index for the event
    @param category   A guild history sub-category for the history event type
--]]
function GrayMarket.OnResponseReceived(event, guildId, category)
  local numEvents = GetNumGuildEvents(guildId, GUILD_HISTORY_STORE)
  local guildName = GetGuildName(guildId)

  if not GrayMarket.guildPasses[guildName] then
    GrayMarket.guildPasses[guildName] = 1
  end

  if numEvents == 0 or GrayMarket.guildPasses[guildName] == 1 then
    -- initiate older event lookup
    zo_callLater(function() GrayMarket:GetSalesHistory(guildId, false) end, CALL_DELAY)
    return
  end

  -- walk available number of events for successful responses with data
  for i = 1, numEvents do
    local eventType, secsSinceEvent, seller, buyer, qty, itemLink, price = GetGuildEventInfo(guildId, GUILD_HISTORY_STORE, i)
    local tTime = GetTimeStamp() - secsSinceEvent
    local itemId = GrayMarket:GetItemId(itemLink)

    if tTime > GrayMarket.data[LAST_FETCH] and eventType == GUILD_EVENT_ITEM_SOLD and itemId ~= nil then
      local guildData = GrayMarket.data[GUILD_LIST][guildName]

      -- track sales by guild, item, and if sold through a vendor or to a guild member
      if not guildData[GUILD_SALE_LIST] then
        guildData[GUILD_SALE_LIST] = {
          [MEMBER_SALE] = {},
          [VENDOR_SALE] = {}
        }
      end

      -- individual item
      local item = {
        [TRANSACTION_TIME] = tTime,
        ["buyer"] = buyer,
        ["seller"] = seller,
        [ITEM_QUANTITY] = qty,
        [ITEM_PRICE] = price
      }

      -- save as vendor transaction or sale to a guild member internally
      if not guildData[GUILD_MEMBERS][buyer] then
        GrayMarket:InitItem(guildName, VENDOR_SALE, itemId)
        table.insert(guildData[GUILD_SALE_LIST][VENDOR_SALE][itemId], item)

        GrayMarket.newSales = GrayMarket.newSales + 1
      else
        GrayMarket:InitItem(guildName, MEMBER_SALE, itemId)
        table.insert(guildData[GUILD_SALE_LIST][MEMBER_SALE][itemId], item)

        GrayMarket.newSales = GrayMarket.newSales + 1
      end
    end
  end
end

--[[
  OnMouseDown(eventCode, button, ctrl, alt, shift, command):

    Hook to capture mouse click event .. if an item link has been set, due
    to hover, then capture shift + click and link item to chat

    @param eventCode    Event code triggered
    @param button       Not sure, but I think if hovering a control like editbox
    @param ctrl         Has ctrl button been pressed as well
    @param alt          Has alt button been pressed as well
    @param shift        Has shift button been pressed as well
    @param command      Not sure, comand button?  On a Mac?  /shrug
]]--
function GrayMarket.OnMouseDown(eventCode, button, ctrl, alt, shift, command)
  if button == 1 and shift and GrayMarket.itemLink then
    GrayMarket:LinkToChat()
  end
end

--[[
  PerformManualFetch():

    Used in keybind to easily trigger a full fetch and purge cycle
]]--
function PerformManualFetch()
  GrayMarket:PerformFullPurgeAndFetch()
end

-- event registration for add-on
EVENT_MANAGER:RegisterForEvent(GrayMarket.name, EVENT_ADD_ON_LOADED, GrayMarket.OnAddOnLoaded)
EVENT_MANAGER:RegisterForEvent(GrayMarket.name, EVENT_GLOBAL_MOUSE_DOWN, GrayMarket.OnMouseDown)
